import os.path
import sqlite3

from api.database.database import (PreparedStatement, FilledStatement,
                                   AbstractTransaction, ReadTransaction, WriteTransaction,
                                   DbConnection,
                                   DbConnectionPool,
                                   TransactionIsolationLevel,
                                   DbResult, DbResultSet, DbResultRow, DbValueType, DatabaseWarning,
                                   DB_RESULT_EXCEPTION, DB_RESULT_WARNINGS, DB_RESULT_SET, DB_RESULT_AFFECTED_ROWS)


def __create_pool() -> DbConnectionPool:
    import api
    if api.config["DB_ENGINE"] == "mysql":
        print("Using MySQL engine for database")
        import api.database.mysql_connector
        from api.miscellaneous import DEBUG_ENABLED
        
        db_config = api.config["MYSQL"]
        __factory = mysql_connector.MySqlDbConnectionFactory(
            not DEBUG_ENABLED,
            DEBUG_ENABLED,
            db_config.get("host", None),
            db_config.get("port", 3306),
            db_config.get("unix_socket", None),
            db_config["user"],
            db_config["password"],
            db_config["database"]
        )
    elif api.config["DB_ENGINE"] == "postgres":
        print("Using PostgreSQL engine for database")
        import api.database.postgres_connector
        from api.miscellaneous import DEBUG_ENABLED
    
        db_config = api.config["POSTGRES"]
        __factory = postgres_connector.PostgresDbConnectionFactory(
            not DEBUG_ENABLED,
            DEBUG_ENABLED,
            db_config["host"],
            db_config.get("port", 5432),
            db_config["user"],
            db_config["password"],
            db_config["database"]
        )
    elif api.config["DB_ENGINE"] == "sqlite":
        print("Using SQLite engine for database")
        import api.database.sqlite_connector
        from pathlib import Path
        
        db_config = api.config["SQLITE"]
        db_path = Path(os.path.join(os.getcwd(), db_config["file"]))
        print(f"SQLite file: {db_path}")
        
        if db_config.get("init_schema", False) or db_config.get("init_data", False):
            file_existed = db_path.exists()
            db = sqlite3.connect(db_path)
            cur = db.cursor()
            if db_config.get("init_schema", False):
                print("Initializing sqlite schema")
                cur.executescript(Path(os.path.join(os.getcwd(), db_config["schema"])).read_text(encoding="UTF-8"))
            if db_config.get("init_data", False) and not file_existed:
                print("Initializing sqlite data")
                cur.executescript(Path(os.path.join(os.getcwd(), api.config["DB_DATA"])).read_text(encoding="UTF-8"))
            db.commit()
            db.close()
        
        __factory = sqlite_connector.SqLiteDbConnectionFactory(db_path)
    else:  # pragma: no cover
        raise ValueError(f"Unknown database engine: {api.config['DB_ENGINE']}")
    
    connection_config = api.config["DB_CONNECTIONS"]
    return DbConnectionPool(
        __factory,
        connection_config["max_count"],
        connection_config["readonly_percent"],
        connection_config["max_wait_time_sec"],
        connection_config["max_waiting_count"],
        connection_config["max_read_attempts"],
        connection_config["max_write_attempts"]
    )


db_pool = __create_pool()
