import math
import time
import ipaddress
from ipaddress import IPv4Address
from threading import Lock

from api.miscellaneous.diagnostics import DIAGNOSTICS_TRACKER, DiagnosticsDataProvider


def _get_truncated_address(ip_string: str) -> bytes:
    parsed = ipaddress.ip_address(ip_string)
    if isinstance(parsed, IPv4Address):
        return parsed.packed[1:4]
    else:
        return parsed.packed[2:8]


class IntervalRateLimiter:
    
    def __init__(self, window_size_ns: int, max_request_count: int, interval_length_ns: int, diagnostics_key: str or None):
        super().__init__()
        if interval_length_ns < 1:
            raise ValueError("Interval length must be greater than 0")  # pragma: no cover
        if max_request_count < 1:
            raise ValueError("Request count must be greater than 0")  # pragma: no cover
        if window_size_ns <= interval_length_ns:
            raise ValueError("Window size must be greater than interval length")  # pragma: no cover
        self._lock = Lock()
        self._interval_length_ns = interval_length_ns
        self._window_size_ns = window_size_ns
        self._max_request_count = max_request_count
        self._latest_count: list[tuple[int, int]] = []
        self._current_request_count = 0
        
        self._diagnostics_reject_counter = None
        if diagnostics_key is not None:
            self._diagnostics_reject_counter = DIAGNOSTICS_TRACKER.register_counter(f"{diagnostics_key}.rejections")
    
    def check_new_request(self) -> bool:
        current_time = time.time_ns()
        is_allowed = self._check_new_request_with_time(current_time)
        if not is_allowed and self._diagnostics_reject_counter is not None:
            self._diagnostics_reject_counter.trigger()
        return is_allowed
    
    def _check_new_request_with_time(self, current_time_ns: int):
        with self._lock:
            if len(self._latest_count) == 0:
                self._current_request_count = 1
                self._latest_count.append((current_time_ns, 1))
                return True
            
            last_time, last_count = self._latest_count[0]
            if current_time_ns < last_time + self._interval_length_ns:
                if self._current_request_count >= self._max_request_count:
                    return False
                self._latest_count[0] = (last_time, last_count + 1)
                self._current_request_count += 1
                return True
            
            oldest_time, oldest_count = self._latest_count[-1]
            while oldest_time < current_time_ns - self._window_size_ns:
                self._current_request_count -= oldest_count
                assert self._current_request_count >= 0
                del self._latest_count[-1]
                if len(self._latest_count) == 0:
                    break
                oldest_time, oldest_count = self._latest_count[-1]
            
            if self._current_request_count >= self._max_request_count:
                return False
            
            self._latest_count.insert(0, (current_time_ns, 1))
            self._current_request_count += 1
            return True


class HostBasedCounterRateLimiter(DiagnosticsDataProvider):
    
    def __init__(self, window_size_ns: int, max_request_count: int, diagnostics_key: str or None):
        super().__init__()
        self._lock = Lock()
        if max_request_count < 1:
            raise ValueError("Request count must be greater than 0")  # pragma: no cover
        if window_size_ns <= 1:
            raise ValueError("Window size must be positive")  # pragma: no cover
        if window_size_ns > 4 * 60 * 60 * 1000 * 1000 * 1000:
            # For everything below, the dictionary size should stay in reasonable limits
            # 100 requests per second for four hours from unique truncated ip's is about 1 million entries
            raise ValueError("Window size must not be greater than 4 hours")  # pragma: no cover
        self._window_size_ns = window_size_ns
        self._next_window_start = math.ceil(time.time_ns() / window_size_ns) * window_size_ns
        self._max_request_count = max_request_count
        self._counter: dict[bytes, int] = {}
        
        self._diagnostics_unique_limited_hosts = None
        if diagnostics_key is not None:
            self._diagnostics_limited_requests = 0
            self._diagnostics_unique_limited_hosts: dict[bytes, None] = {}
            DIAGNOSTICS_TRACKER.register_provider(diagnostics_key, self)
    
    def check_new_request(self, ip_string: str) -> bool:
        with self._lock:
            current_time = time.time_ns()
            if current_time >= self._next_window_start:
                self._counter = {}
                self._next_window_start = math.ceil(current_time / self._window_size_ns) * self._window_size_ns
            address = _get_truncated_address(ip_string)
            count = self._counter.get(address, 0)
            if count >= self._max_request_count:
                if self._diagnostics_limited_requests is not None:
                    self._diagnostics_unique_limited_hosts[address] = None
                    self._diagnostics_limited_requests += 1
                return False
            self._counter[address] = count + 1
            return True
    
    def get_diagnostics_data(self) -> dict[str, int or str]:
        with self._lock:
            res = {
                "unique_limited_hosts": len(self._diagnostics_unique_limited_hosts),
                "limited_requests": self._diagnostics_limited_requests
            }
            self._diagnostics_unique_limited_hosts = {}
            self._diagnostics_limited_requests = 0
            return res


def create_configured_host_rate_limiters(id: str, config: list[dict]) -> tuple[HostBasedCounterRateLimiter, ...]:
    return tuple([
        HostBasedCounterRateLimiter(
            entry["window_size_seconds"] * 1000 * 1000 * 1000,
            entry["max_request_count"],
            f"rate_limiter.{id}.{entry['id']}"
        )
        for entry in config
    ])
